@@ -1,7 +1,7 @@
 /*
  * mdadm - manage Linux "md" devices aka RAID arrays.
  *
- * Copyright (C) 2001-2002 Neil Brown <neilb@cse.unsw.edu.au>
+ * Copyright (C) 2001-2006 Neil Brown <neilb@suse.de>
  *
  *
  *    This program is free software; you can redistribute it and/or modify
@@ -33,6 +33,8 @@
 #include	<glob.h>
 #include	<fnmatch.h>
 #include	<ctype.h>
+#include	<pwd.h>
+#include	<grp.h>
 
 /*
  * Read the config file
@@ -70,9 +72,24 @@
 #ifndef CONFFILE
 #define CONFFILE "/etc/mdadm.conf"
 #endif
+#ifndef CONFFILE2
+/* for Debian compatibility .... */
+#define CONFFILE2 "/etc/mdadm/mdadm.conf"
+#endif
 char DefaultConfFile[] = CONFFILE;
+char DefaultAltConfFile[] = CONFFILE2;
 
-char *keywords[] = { "device", "array", "mailaddr", "program", NULL };
+enum linetype { Devices, Array, Mailaddr, Mailfrom, Program, CreateDev, Homehost, LTEnd };
+char *keywords[] = {
+	[Devices]  = "devices",
+	[Array]    = "array",
+	[Mailaddr] = "mailaddr",
+	[Mailfrom] = "mailfrom",
+	[Program]  = "program",
+	[CreateDev]   = "create",
+	[Homehost] = "homehost",
+	[LTEnd]    = NULL
+};
 
 /*
  * match_keyword returns an index into the keywords array, or -1 for no match
@@ -202,17 +219,20 @@ struct conf_dev {
     char *name;
 } *cdevlist = NULL;
 
-void load_partitions(void)
+mddev_dev_t load_partitions(void)
 {
 	FILE *f = fopen("/proc/partitions", "r");
 	char buf[1024];
+	mddev_dev_t rv = NULL;
 	if (f == NULL) {
 		fprintf(stderr, Name ": cannot open /proc/partitions\n");
-		return;
+		return NULL;
 	}
 	while (fgets(buf, 1024, f)) {
 		int major, minor;
 		char *name, *mp;
+		mddev_dev_t d;
+
 		buf[1023] = '\0';
 		if (buf[0] != ' ')
 			continue;
@@ -221,19 +241,129 @@ void load_partitions(void)
 			continue;
 		minor = strtoul(mp, NULL, 10);
 
-		name = map_dev(major, minor);
-		if (name) {
-			struct conf_dev *cd;
+		name = map_dev(major, minor, 1);
 
-			cd = malloc(sizeof(*cd));
-			cd->name = strdup(name);
-			cd->next = cdevlist;
-			cdevlist = cd;
-		}
+		d = malloc(sizeof(*d));
+		d->devname = strdup(name);
+		d->next = rv;
+		d->used = 0;
+		rv = d;
 	}
 	fclose(f);
+	return rv;
 }
 
+struct createinfo createinfo = {
+#ifdef DEBIAN
+	.gid = 6, /* disk */
+	.mode = 0660,
+#else
+	.mode = 0600,
+#endif
+};
+
+int parse_auto(char *str, char *msg, int config)
+{
+	int autof;
+	if (str == NULL || *str == 0)
+		autof = 2;
+	else if (strcasecmp(str,"no")==0)
+		autof = 1;
+	else if (strcasecmp(str,"yes")==0)
+		autof = 2;
+	else if (strcasecmp(str,"md")==0)
+		autof = config?5:3;
+	else {
+		/* There might be digits, and maybe a hypen, at the end */
+		char *e = str + strlen(str);
+		int num = 4;
+		int len;
+		while (e > str && isdigit(e[-1]))
+			e--;
+		if (*e) {
+			num = atoi(e);
+			if (num <= 0) num = 1;
+		}
+		if (e > str && e[-1] == '-')
+			e--;
+		len = e - str;
+		if ((len == 2 && strncasecmp(str,"md",2)==0)) {
+			autof = config ? 5 : 3;
+		} else if ((len == 3 && strncasecmp(str,"mdp",3)==0)) {
+			autof = config ? 6 : 4;
+		} else if ((len == 1 && strncasecmp(str,"p",1)==0) ||
+			   (len >= 4 && strncasecmp(str,"part",4)==0)) {
+			autof = 6;
+		} else {
+			fprintf(stderr, Name ": %s arg of \"%s\" unrecognised: use no,yes,md,mdp,part\n"
+				"        optionally followed by a number.\n",
+				msg, str);
+			exit(2);
+		}
+		autof |= num << 3;
+	}
+	return autof;
+}
+
+static void createline(char *line)
+{
+	char *w;
+	char *ep;
+
+	for (w=dl_next(line); w!=line; w=dl_next(w)) {
+		if (strncasecmp(w, "auto=", 5) == 0)
+			createinfo.autof = parse_auto(w+5, "auto=", 1);
+		else if (strncasecmp(w, "owner=", 6) == 0) {
+			if (w[6] == 0) {
+				fprintf(stderr, Name ": missing owner name\n");
+				continue;
+			}
+			createinfo.uid = strtoul(w+6, &ep, 10);
+			if (*ep != 0) {
+#ifndef STATIC
+				struct passwd *pw;
+				/* must be a name */
+				pw = getpwnam(w+6);
+				if (pw)
+					createinfo.uid = pw->pw_uid;
+				else
+#endif /* STATIC */
+					fprintf(stderr, Name ": CREATE user %s not found\n", w+6);
+			}
+		} else if (strncasecmp(w, "group=", 6) == 0) {
+			if (w[6] == 0) {
+				fprintf(stderr, Name ": missing group name\n");
+				continue;
+			}
+			createinfo.gid = strtoul(w+6, &ep, 10);
+			if (*ep != 0) {
+#ifndef STATIC
+				struct group *gr;
+				/* must be a name */
+				gr = getgrnam(w+6);
+				if (gr)
+					createinfo.gid = gr->gr_gid;
+				else
+#endif /* STATIC */
+					fprintf(stderr, Name ": CREATE group %s not found\n", w+6);
+			}
+		} else if (strncasecmp(w, "mode=", 5) == 0) {
+			if (w[5] == 0) {
+				fprintf(stderr, Name ": missing CREATE mode\n");
+				continue;
+			}
+			createinfo.mode = strtoul(w+5, &ep, 8);
+			if (*ep != 0) {
+				createinfo.mode = 0600;
+				fprintf(stderr, Name ": unrecognised CREATE mode %s\n",
+					w+5);
+			}
+		} else {
+			fprintf(stderr, Name ": unrecognised word on CREATE line: %s\n",
+				w);
+		}
+	}
+}
 
 void devline(char *line) 
 {
@@ -241,14 +371,11 @@ void devline(char *line)
 	struct conf_dev *cd;
 
 	for (w=dl_next(line); w != line; w=dl_next(w)) {
-		if (w[0] == '/') {
+		if (w[0] == '/' || strcasecmp(w, "partitions") == 0) {
 			cd = malloc(sizeof(*cd));
 			cd->name = strdup(w);
 			cd->next = cdevlist;
 			cdevlist = cd;
-		} else if (strcasecmp(w, "partitions") == 0) {
-			/* read /proc/partitions, and look major/minor up in /dev */
-			load_partitions();
 		} else {
 			fprintf(stderr, Name ": unreconised word on DEVICE line: %s\n",
 				w);
@@ -270,11 +397,16 @@ void arrayline(char *line)
 	mis.super_minor = UnSet;
 	mis.level = UnSet;
 	mis.raid_disks = UnSet;
-	mis.spare_disks = UnSet;
+	mis.spare_disks = 0;
 	mis.devices = NULL;
 	mis.devname = NULL;
 	mis.spare_group = NULL;
 	mis.autof = 0;
+	mis.next = NULL;
+	mis.st = NULL;
+	mis.bitmap_fd = -1;
+	mis.bitmap_file = NULL;
+	mis.name[0] = 0;
 
 	for (w=dl_next(line); w!=line; w=dl_next(w)) {
 		if (w[0] == '/') {
@@ -293,7 +425,7 @@ void arrayline(char *line)
 					fprintf(stderr, Name ": bad uuid: %s\n", w);
 			}
 		} else if (strncasecmp(w, "super-minor=", 12)==0 ) {
-			if (mis.super_minor >= 0)
+			if (mis.super_minor != UnSet)
 				fprintf(stderr, Name ": only specify super-minor once, %s ignored.\n",
 					w);
 			else {
@@ -305,6 +437,22 @@ void arrayline(char *line)
 					mis.super_minor = UnSet;
 				}
 			}
+		} else if (strncasecmp(w, "name=", 5)==0) {
+			if (mis.name[0])
+				fprintf(stderr, Name ": only specify name once, %s ignored.\n",
+					w);
+			else if (strlen(w+5) > 32)
+				fprintf(stderr, Name ": name too long, ignoring %s\n", w);
+			else
+				strcpy(mis.name, w+5);
+
+		} else if (strncasecmp(w, "bitmap=", 7) == 0) {
+			if (mis.bitmap_file)
+				fprintf(stderr, Name ": only specify bitmap file once. %s ignored\n",
+					w);
+			else
+				mis.bitmap_file = w+7;
+
 		} else if (strncasecmp(w, "devices=", 8 ) == 0 ) {
 			if (mis.devices)
 				fprintf(stderr, Name ": only specify devices once (use a comma separated list). %s ignored\n",
@@ -340,32 +488,7 @@ void arrayline(char *line)
 				fprintf(stderr, Name ": metadata format %s unknown, ignored.\n", w+9);
 		} else if (strncasecmp(w, "auto=", 5) == 0 ) {
 			/* whether to create device special files as needed */
-			if (strcasecmp(w+5, "no")==0)
-				mis.autof = 0;
-			else if (strcasecmp(w+5,"yes")==0 || strcasecmp(w+5,"md")==0)
-				mis.autof = -1;
-			else {
-				/* There might be digits, and maybe a hypen, at the end */
-				char *e = w+5 + strlen(w+5);
-				int num = 4;
-				int len;
-				while (e > w+5 && isdigit(e[-1]))
-					e--;
-				if (*e) {
-					num = atoi(e);
-					if (num <= 0) num = 1;
-				}
-				if (e > w+5 && e[-1] == '-')
-					e--;
-				len = e - (w+5);
-				if ((len == 3 && strncasecmp(w+5,"mdp",3)==0) ||
-				    (len == 1 && strncasecmp(w+5,"p",1)==0) ||
-				    (len >= 4 && strncasecmp(w+5,"part",4)==0))
-					mis.autof = num;
-				else 
-					fprintf(stderr, Name ": auto type of \"%s\" ignored for %s\n",
-						w+5, mis.devname?mis.devname:"unlabeled-array");
-			}
+			mis.autof = parse_auto(w+5, "auto type", 0);
 		} else {
 			fprintf(stderr, Name ": unrecognised word on ARRAY line: %s\n",
 				w);
@@ -373,7 +496,7 @@ void arrayline(char *line)
 	}
 	if (mis.devname == NULL)
 		fprintf(stderr, Name ": ARRAY line with no device\n");
-	else if (mis.uuid_set == 0 && mis.devices == NULL && mis.super_minor == UnSet)
+	else if (mis.uuid_set == 0 && mis.devices == NULL && mis.super_minor == UnSet && mis.name[0] == 0)
 		fprintf(stderr, Name ": ARRAY line %s has no identity information.\n", mis.devname);
 	else {
 		mi = malloc(sizeof(*mi));
@@ -399,6 +522,23 @@ void mailline(char *line)
 	}
 }
 
+static char *alert_mail_from = NULL;
+void mailfromline(char *line)
+{
+	char *w;
+
+	for (w=dl_next(line); w != line ; w=dl_next(w)) {
+		if (alert_mail_from == NULL)
+			alert_mail_from = strdup(w);
+		else {
+			char *t= NULL;
+			asprintf(&t, "%s %s", alert_mail_from, w);
+			free(alert_mail_from);
+			alert_mail_from = t;
+		}
+	}
+}
+
 
 static char *alert_program = NULL;
 void programline(char *line)
@@ -414,6 +554,20 @@ void programline(char *line)
 	}
 }
 
+static char *home_host = NULL;
+void homehostline(char *line)
+{
+	char *w;
+
+	for (w=dl_next(line); w != line ; w=dl_next(w)) {
+		if (home_host == NULL)
+			home_host = strdup(w);
+		else
+			fprintf(stderr, Name ": excess host name on HOMEHOST line: %s - ignored\n",
+				w);
+	}
+}
+
 
 int loaded = 0;
 
@@ -431,29 +585,53 @@ void load_conffile(char *conffile)
 		return;
 	}
 	if (strcmp(conffile, "partitions")==0) {
-		load_partitions();
+		char *list = dl_strdup("DEV");
+		dl_init(list);
+		dl_add(list, dl_strdup("partitions"));
+		devline(list);
+		free_line(list);
 		loaded = 1;
 		return;
 	}
 	f = fopen(conffile, "r");
-	if (f ==NULL)
+	/* Debian chose to relocate mdadm.conf into /etc/mdadm/.
+	 * To allow Debian users to compile from clean source and still
+	 * have a working mdadm, we read /etc/mdadm/mdadm.conf
+	 * if /etc/mdadm.conf doesn't exist
+	 */
+	if (f == NULL &&
+	    conffile == DefaultConfFile) {
+		f = fopen(DefaultAltConfFile, "r");
+		if (f)
+			conffile = DefaultAltConfFile;
+	}
+	if (f == NULL)
 		return;
 
 	loaded = 1;
 	while ((line=conf_line(f))) {
 		switch(match_keyword(line)) {
-		case 0: /* DEVICE */
+		case Devices:
 			devline(line);
 			break;
-		case 1: /* ARRAY */
+		case Array:
 			arrayline(line);
 			break;
-		case 2: /* MAIL */
+		case Mailaddr:
 			mailline(line);
 			break;
-		case 3: /* PROGRAM */
+		case Mailfrom:
+			mailfromline(line);
+			break;
+		case Program:
 			programline(line);
 			break;
+		case CreateDev:
+			createline(line);
+			break;
+		case Homehost:
+			homehostline(line);
+			break;
 		default:
 			fprintf(stderr, Name ": Unknown keyword %s\n", line);
 		}
@@ -471,12 +649,29 @@ char *conf_get_mailaddr(char *conffile)
 	return alert_email;
 }
 
+char *conf_get_mailfrom(char *conffile)
+{
+	load_conffile(conffile);
+	return alert_mail_from;
+}
+
 char *conf_get_program(char *conffile)
 {
 	load_conffile(conffile);
 	return alert_program;
 }
 
+char *conf_get_homehost(char *conffile)
+{
+	load_conffile(conffile);
+	return home_host;
+}
+
+struct createinfo *conf_get_create_info(char *conffile)
+{
+	load_conffile(conffile);
+	return &createinfo;
+}
 
 mddev_ident_t conf_get_ident(char *conffile, char *dev)
 {
@@ -504,16 +699,25 @@ mddev_dev_t conf_get_devs(char *conffile)
 	}
     
 	load_conffile(conffile);
-    
+
+	if (cdevlist == NULL)
+		/* default to 'partitions */
+		dlist = load_partitions();
+
 	for (cd=cdevlist; cd; cd=cd->next) {
-		glob(cd->name, flags, NULL, &globbuf);
-		flags |= GLOB_APPEND;
+		if (strcasecmp(cd->name, "partitions")==0 && dlist == NULL)
+			dlist = load_partitions();
+		else {
+			glob(cd->name, flags, NULL, &globbuf);
+			flags |= GLOB_APPEND;
+		}
 	}
 	if (flags & GLOB_APPEND) {
 		for (i=0; i<globbuf.gl_pathc; i++) {
 			mddev_dev_t t = malloc(sizeof(*t));
 			t->devname = strdup(globbuf.gl_pathv[i]);
 			t->next = dlist;
+			t->used = 0;
 			dlist = t;
 /*	printf("one dev is %s\n", t->devname);*/
 		}
