@@ -1,7 +1,7 @@
 /*
  * mdadm - manage Linux "md" devices aka RAID arrays.
  *
- * Copyright (C) 2001-2004 Neil Brown <neilb@cse.unsw.edu.au>
+ * Copyright (C) 2001-2006 Neil Brown <neilb@suse.de>
  *
  *
  *    This program is free software; you can redistribute it and/or modify
@@ -28,7 +28,11 @@
  */
 
 #include "mdadm.h"
-
+#ifndef UCLIBC
+#include <openssl/sha.h> /* for SHA1 */
+#else
+extern unsigned char *SHA1(unsigned char *buf, int len, unsigned char *dest);
+#endif
 /*
  * All handling for the 0.90.0 version superblock is in
  * this file.
@@ -84,7 +88,7 @@ void super0_swap_endian(struct mdp_superblock_s *sb)
 
 #ifndef MDASSEMBLE
 
-static void examine_super0(void *sbv)
+static void examine_super0(void *sbv, char *homehost)
 {
 	mdp_super_t *sb = sbv;
 	time_t atime;
@@ -94,10 +98,18 @@ static void examine_super0(void *sbv)
 	printf("          Magic : %08x\n", sb->md_magic);
 	printf("        Version : %02d.%02d.%02d\n", sb->major_version, sb->minor_version,
 	       sb->patch_version);
-	if (sb->minor_version >= 90)
-		printf("           UUID : %08x:%08x:%08x:%08x\n", sb->set_uuid0, sb->set_uuid1,
+	if (sb->minor_version >= 90) {
+		printf("           UUID : %08x:%08x:%08x:%08x", sb->set_uuid0, sb->set_uuid1,
 		       sb->set_uuid2, sb->set_uuid3);
-	else
+		if (homehost) {
+			unsigned char *hash = SHA1((unsigned char *)homehost,
+						   strlen(homehost),
+						   NULL);
+			if (memcmp(&sb->set_uuid2, hash, 8)==0)
+				printf(" (local to host %s)", homehost);
+		}
+		printf("\n");
+	} else
 		printf("           UUID : %08x\n", sb->set_uuid0);
 
 	atime = sb->ctime;
@@ -141,8 +153,10 @@ static void examine_super0(void *sbv)
 				printf("     New Layout : %s\n", c?c:"-unknown-");
 			}
 			if (sb->level == 10) {
-				printf("     New Layout : near=%d, far=%d\n",
-				       sb->new_layout&255, (sb->new_layout>>8)&255);
+				printf("     New Layout : near=%d, %s=%d\n",
+				       sb->new_layout&255,
+				       (sb->new_layout&0x10000)?"offset":"far",
+				       (sb->new_layout>>8)&255);
 			}
 		}
 		if (sb->new_chunk != sb->chunk_size)
@@ -170,8 +184,10 @@ static void examine_super0(void *sbv)
 		printf("         Layout : %s\n", c?c:"-unknown-");
 	}
 	if (sb->level == 10) {
-		printf("         Layout : near=%d, far=%d\n",
-		       sb->layout&255, (sb->layout>>8)&255);
+		printf("         Layout : near=%d, %s=%d\n",
+		       sb->layout&255,
+		       (sb->layout&0x10000)?"offset":"far",
+		       (sb->layout>>8)&255);
 	}
 	switch(sb->level) {
 	case 0:
@@ -215,9 +231,17 @@ static void brief_examine_super0(void *sbv)
 {
 	mdp_super_t *sb = sbv;
 	char *c=map_num(pers, sb->level);
+	char devname[20];
+	struct stat stb;
+
+	sprintf(devname, "/dev/md%d", sb->md_minor);
+	if (stat(devname, &stb) != 0) {
+		/* /dev/mdX doesn't exist, so use /dev/md/X */
+		sprintf(devname, "/dev/md/%d", sb->md_minor);
+	}
 
 	printf("ARRAY %s level=%s num-devices=%d UUID=",
-	       get_md_name(sb->md_minor),
+	       devname,
 	       c?c:"-unknown-", sb->raid_disks);
 	if (sb->minor_version >= 90)
 		printf("%08x:%08x:%08x:%08x", sb->set_uuid0, sb->set_uuid1,
@@ -227,7 +251,7 @@ static void brief_examine_super0(void *sbv)
 	printf("\n");
 }
 
-static void detail_super0(void *sbv)
+static void detail_super0(void *sbv, char *homehost)
 {
 	mdp_super_t *sb = sbv;
 	printf("           UUID : ");
@@ -236,6 +260,13 @@ static void detail_super0(void *sbv)
 		       sb->set_uuid2, sb->set_uuid3);
 	else
 		printf("%08x", sb->set_uuid0);
+	if (homehost) {
+		unsigned char *hash = SHA1((unsigned char *)homehost,
+					   strlen(homehost),
+					   NULL);
+		if (memcmp(&sb->set_uuid2, hash, 8)==0)
+			printf(" (local to host %s)", homehost);
+	}
 	printf("\n         Events : %d.%d\n\n", sb->events_hi, sb->events_lo);
 }
 
@@ -250,6 +281,17 @@ static void brief_detail_super0(void *sbv)
 		printf("%08x", sb->set_uuid0);
 }
 #endif
+
+static int match_home0(void *sbv, char *homehost)
+{
+	mdp_super_t *sb = sbv;
+	unsigned char *hash = SHA1((unsigned char *)homehost,
+				   strlen(homehost),
+				   NULL);
+
+	return (memcmp(&sb->set_uuid2, hash, 8)==0);
+}
+
 static void uuid_from_super0(int uuid[4], void * sbv)
 {
 	mdp_super_t *super = sbv;
@@ -304,7 +346,7 @@ static void getinfo_super0(struct mdinfo *info, void *sbv)
 	} else
 		info->reshape_active = 0;
 
-	info->name[0] = 0;
+	sprintf(info->name, "%d", sb->md_minor);
 	/* work_disks is calculated rather than read directly */
 	for (i=0; i < MD_SB_DISKS; i++)
 		if ((sb->disks[i].state & (1<<MD_DISK_SYNC)) &&
@@ -315,8 +357,13 @@ static void getinfo_super0(struct mdinfo *info, void *sbv)
 }
 
 
-static int update_super0(struct mdinfo *info, void *sbv, char *update, char *devname, int verbose)
+static int update_super0(struct mdinfo *info, void *sbv, char *update,
+			 char *devname, int verbose,
+			 int uuid_set, char *homehost)
 {
+	/* NOTE: for 'assemble' and 'force' we need to return non-zero if any change was made.
+	 * For others, the return value is ignored.
+	 */
 	int rv = 0;
 	mdp_super_t *sb = sbv;
 	if (strcmp(update, "sparc2.2")==0 ) {
@@ -367,18 +414,25 @@ static int update_super0(struct mdinfo *info, void *sbv, char *update, char *dev
 				sb->disks[i].state = 0;
 	}
 	if (strcmp(update, "force")==0) {
+		__u32 ehi = sb->events_hi, elo = sb->events_lo;
 		sb->events_hi = (info->events>>32) & 0xFFFFFFFF;
 		sb->events_lo = (info->events) & 0xFFFFFFFF;
-		if (sb->level == 5 || sb->level == 4 || sb->level == 6)
+		if (sb->events_hi != ehi ||
+		    sb->events_lo != elo)
+			rv = 1;
+		if ((sb->level == 5 || sb->level == 4 || sb->level == 6) &&
+		    (sb->state & (1 << MD_SB_CLEAN)) == 0) {
 			/* need to force clean */
 			sb->state |= (1 << MD_SB_CLEAN);
+			rv = 1;
+		}
 	}
 	if (strcmp(update, "assemble")==0) {
 		int d = info->disk.number;
 		int wonly = sb->disks[d].state & (1<<MD_DISK_WRITEMOSTLY);
-		sb->disks[d].state &= ~(1<<MD_DISK_WRITEMOSTLY);
-		if (sb->disks[d].state != info->disk.state) {
-			sb->disks[d].state = info->disk.state & wonly;
+		if ((sb->disks[d].state & ~(1<<MD_DISK_WRITEMOSTLY))
+		    != info->disk.state) {
+			sb->disks[d].state = info->disk.state | wonly;
 			rv = 1;
 		}
 	}
@@ -411,11 +465,29 @@ static int update_super0(struct mdinfo *info, void *sbv, char *update, char *dev
 		sb->state &= ~(1<<MD_SB_CLEAN);
 		sb->recovery_cp = 0;
 	}
+	if (strcmp(update, "homehost") == 0 &&
+	    homehost) {
+		uuid_set = 0;
+		update = "uuid";
+		info->uuid[0] = sb->set_uuid0;
+		info->uuid[1] = sb->set_uuid1;
+	}
 	if (strcmp(update, "uuid") == 0) {
+		if (!uuid_set && homehost) {
+			unsigned char *hash = SHA1((unsigned char*)homehost,
+						   strlen(homehost),
+						   NULL);
+			memcpy(info->uuid+2, hash, 8);
+		}
 		sb->set_uuid0 = info->uuid[0];
 		sb->set_uuid1 = info->uuid[1];
 		sb->set_uuid2 = info->uuid[2];
 		sb->set_uuid3 = info->uuid[3];
+		if (sb->state & (1<<MD_SB_BITMAP_PRESENT)) {
+			struct bitmap_super_s *bm;
+			bm = (struct bitmap_super_s*)(sb+1);
+			uuid_from_super0((int*)bm->uuid, sbv);
+		}
 	}
 	if (strcmp(update, "_reshape_progress")==0)
 		sb->reshape_position = info->reshape_progress;
@@ -430,9 +502,17 @@ static __u64 event_super0(void *sbv)
 	return md_event(sb);
 }
 
+/*
+ * For verion-0 superblock, the homehost is 'stored' in the
+ * uuid.  8 bytes for a hash of the host leaving 8 bytes
+ * of random material.
+ * We use the first 8 bytes (64bits) of the sha1 of the
+ * host name
+ */
 
 
-static int init_super0(struct supertype *st, void **sbp, mdu_array_info_t *info, unsigned long long size, char *ignored_name)
+static int init_super0(struct supertype *st, void **sbp, mdu_array_info_t *info,
+		       unsigned long long size, char *ignored_name, char *homehost)
 {
 	mdp_super_t *sb = malloc(MD_SB_BYTES + sizeof(bitmap_super_t));
 	int spares;
@@ -441,6 +521,7 @@ static int init_super0(struct supertype *st, void **sbp, mdu_array_info_t *info,
 
 	if (info->major_version == -1) {
 		/* zeroing the superblock */
+		*sbp = sb;
 		return 0;
 	}
 
@@ -475,6 +556,12 @@ static int init_super0(struct supertype *st, void **sbp, mdu_array_info_t *info,
 	}
 	if (rfd >= 0)
 		close(rfd);
+	if (homehost) {
+		unsigned char *hash = SHA1((unsigned char*)homehost,
+					   strlen(homehost),
+					   NULL);
+		memcpy(&sb->set_uuid2, hash, 8);
+	}
 
 	sb->utime = sb->ctime;
 	sb->state = info->state;
@@ -535,6 +622,12 @@ static int store_super0(struct supertype *st, int fd, void *sbv)
 	if (write(fd, super, sizeof(*super)) != sizeof(*super))
 		return 4;
 
+	if (super->state & (1<<MD_SB_BITMAP_PRESENT)) {
+		struct bitmap_super_s * bm = (struct bitmap_super_s*)(super+1);
+		if (__le32_to_cpu(bm->magic) == BITMAP_MAGIC)
+			write(fd, bm, sizeof(*bm));
+	}
+
 	fsync(fd);
 	return 0;
 }
@@ -582,8 +675,8 @@ static int compare_super0(void **firstp, void *secondv)
 	if (second->md_magic != MD_SB_MAGIC)
 		return 1;
 	if (!first) {
-		first = malloc(MD_SB_BYTES);
-		memcpy(first, second, MD_SB_BYTES);
+		first = malloc(MD_SB_BYTES + sizeof(struct bitmap_super_s));
+		memcpy(first, second, MD_SB_BYTES + sizeof(struct bitmap_super_s));
 		*firstp = first;
 		return 0;
 	}
@@ -618,6 +711,8 @@ static int load_super0(struct supertype *st, int fd, void **sbp, char *devname)
 	unsigned long long dsize;
 	unsigned long long offset;
 	mdp_super_t *super;
+	int uuid[4];
+	struct bitmap_super_s *bsb;
     
 #ifdef BLKGETSIZE64
 	if (ioctl(fd, BLKGETSIZE64, &dsize) != 0)
@@ -687,6 +782,27 @@ static int load_super0(struct supertype *st, int fd, void **sbp, char *devname)
 		st->max_devs = MD_SB_DISKS;
 	}
 
+	/* Now check on the bitmap superblock */
+	if ((super->state & (1<<MD_SB_BITMAP_PRESENT)) == 0)
+		return 0;
+	/* Read the bitmap superblock and make sure it looks
+	 * valid.  If it doesn't clear the bit.  An --assemble --force
+	 * should get that written out.
+	 */
+	if (read(fd, super+1, sizeof(struct bitmap_super_s))
+	    != sizeof(struct bitmap_super_s))
+		goto no_bitmap;
+
+	uuid_from_super0(uuid, super);
+	bsb = (struct bitmap_super_s *)(super+1);
+	if (__le32_to_cpu(bsb->magic) != BITMAP_MAGIC ||
+	    memcmp(bsb->uuid, uuid, 16) != 0)
+		goto no_bitmap;
+	return 0;
+
+ no_bitmap:
+	super->state &= ~(1<<MD_SB_BITMAP_PRESENT);
+
 	return 0;
 }
 
@@ -725,7 +841,7 @@ static int add_internal_bitmap0(struct supertype *st, void *sbv, int chunk, int
 	 * The bitmap comes immediately after the superblock and must be 60K in size
 	 * at most.  The default size is between 30K and 60K
 	 *
-	 * size is in K,  chunk is in bytes !!!
+	 * size is in sectors,  chunk is in bytes !!!
 	 */
 	unsigned long long bits;
 	unsigned long long max_bits = 60*1024*8;
@@ -735,7 +851,7 @@ static int add_internal_bitmap0(struct supertype *st, void *sbv, int chunk, int
 
 
 	min_chunk = 4096; /* sub-page chunks don't work yet.. */
-	bits = (size * 512)/ min_chunk +1;
+	bits = (size * 512) / min_chunk + 1;
 	while (bits > max_bits) {
 		min_chunk *= 2;
 		bits = (bits+1)/2;
@@ -756,8 +872,6 @@ static int add_internal_bitmap0(struct supertype *st, void *sbv, int chunk, int
 	bms->sync_size = __cpu_to_le64(size);
 	bms->write_behind = __cpu_to_le32(write_behind);
 
-
-
 	return 1;
 }
 
@@ -851,6 +965,7 @@ struct superswitch super0 = {
 	.detail_super = detail_super0,
 	.brief_detail_super = brief_detail_super0,
 #endif
+	.match_home = match_home0,
 	.uuid_from_super = uuid_from_super0,
 	.getinfo_super = getinfo_super0,
 	.update_super = update_super0,
